/*
 * Copyright: GNU Public License 2 applies
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * CDDA2WAV (C) Heiko Eissfeldt heiko@colossus.escape.de
 * parts    (C) Peter Widow
 * parts    (C) Thomas Niederreiter
 * parts    (C) RSA Data Security, Inc.
 *
 * last changes:
 *   18.12.93 - first version,	OK
 *   01.01.94 - generalized & clean up HE
 *   10.06.94 - first linux version HE
 *   12.06.94 - wav header alignment problem fixed HE
 *   12.08.94 - open the cdrom device O_RDONLY makes more sense :-)
 *		no more floating point math
 *		change to sector size 2352 which is more common
 *		sub-q-channel information per kernel ioctl requested
 *		doesn't work as well as before
 *		some new options (-max -i)
 *   29.01.96 - new options for bulk transfer
 *   04.02.96 - security hole fixed
 *   02.01.95 - async i/o via semaphores and shared memory
 *     .03.95 - overlapped reading on sectors
 *     .03.95 - generalized sample rates. all integral divisors are legal
 *     .04.95 - sun format added
 *              more divisors: all integral halves >= 1 allowed
 *		floating point math needed again
 *     .06.95 - bugfix for last track and not d0
 *              tested with photo-cd with audio tracks
 *		tested with xa disk 
 *   01.06.96 - tested with enhanced cd
 *   01.06.96 - tested with cd-plus
 *   02.06.96 - support pipes
 *   02.06.96 - support raw format
 *
 * TODO: 
 *	 debug program for panasonic drives (have none)
 *       more items in file TODO
 */

#define USE_GETOPT_LONG

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <fcntl.h>
#include <time.h>
#include <sys/ioctl.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/wait.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include "share.h"
#include "uti.h"
#include "wav.h"	/* wav file header structures */
#include "sun.h"	/* sun audio file header structures */
#include "raw.h"	/* raw file handling */
#include "interface.h"  /* low level cdrom interfacing */
#include "cdda2wav.h"
#include "resample.h"
#include "indexscan.h"
#include "setuid.h"

#ifdef	ECHO_TO_SOUNDCARD
#include <linux/soundcard.h>
#endif

#ifdef  MD5_SIGNATURES
#include "md5.h"

int md5blocksize, md5count;
MD5_CTX context;
unsigned char MD5_result[16];
#endif

#ifdef USE_GETOPT_LONG
#include <getopt.h>
const char * optstring = "D:I:O:c:b:r:a:t:i:d:o:n:v:M:smweNqxPRB";
struct option options [] = {
	{"device",required_argument,NULL,'D'},
	{"interface",required_argument,NULL,'I'},
	{"output-format",required_argument,NULL,'O'},
	{"channels",required_argument,NULL,'c'},
	{"bits-per-sample",required_argument,NULL,'b'},
	{"rate",required_argument,NULL,'r'},
	{"divider",required_argument,NULL,'r'},
	{"track",required_argument,NULL,'r'},
	{"index",required_argument,NULL,'i'},
	{"duration",required_argument,NULL,'d'},
	{"offset",required_argument,NULL,'o'},
	{"sectors-per-request",required_argument,NULL,'n'},
	{"verbose-level",required_argument,NULL,'v'},
	{"md5",required_argument,NULL,'M'},
	{"stereo",no_argument,NULL,'s'},
	{"mono",no_argument,NULL,'m'},
	{"wait",no_argument,NULL,'w'},
	{"echo",no_argument,NULL,'e'},
	{"no-write",no_argument,NULL,'N'},
	{"quiet",no_argument,NULL,'q'},
	{"max",no_argument,NULL,'x'},
	{"no-overlap",no_argument,NULL,'P'},
	{"dump-rates",no_argument,NULL,'R'},
	{"bulk",no_argument,NULL,'B'},
	{"version",no_argument,NULL,'V'},
	{"help",no_argument,NULL,'h'},
	{0,0,0,0}
};

#endif /* GETOPT_LONG */


/* global variables */
unsigned char dev_name [200] = CD_DEVICE;		/* device name */
unsigned char aux_name [200] = AUX_DEVICE;		/* device name */

int audio = -1;

static int (*InitAudioFile)( int audio, long channels, unsigned long rate,
			    long nBitsPerSample,
			    unsigned long expected_bytes);
static int (*ExitAudioFile)( int audio, unsigned long nBytesDone );
static unsigned long (*GetSndHdrSize)( void );

static unsigned long nSamplesDone = 0;

long cd_fd = -1;
int no_file = 0;

#ifdef	ECHO_TO_SOUNDCARD
long soundcard_fd;
int echo = 0;
#endif

static	int child_pid = 0xff;
int swap;
int OutSampleSize;
int sh_bits;
int channels = CHANNELS;
int SkippedSamples = 0;

static char fname_base[200] = FILENAME;
static long nSamplesToDo;
static long current_track;
static int bulk = 0;

long SamplesNeeded( long amount, long undersampling)
{
  long retval = ((undersampling * 2 + Halved)*amount)/2;
  if (Halved && (nSamplesToDo & 1))
    retval += 2;
  return retval;
}

#ifdef INFOFILES
int write_info_file(char *fname_base, long int track, long int nSamplesDone, int numbered)
{
  FILE *info_fp;
  char fname[200];
  char datetime[30];
  time_t utc_time;

  /* write info file */
  if (!strcmp(fname_base,"standard output")) return 0;
  if (numbered)
    sprintf(fname, "%s.%02ld.info", fname_base, track);
  else
    sprintf(fname, "%s.info", fname_base);
  info_fp = fopen (fname, "w");
  if (!info_fp)
    return -1;

#ifdef MD5_SIGNATURES
  if (md5blocksize)
    MD5Final (MD5_result, &context);
#endif

  utc_time = time(NULL);
  strftime(datetime, sizeof(datetime), "%x %X", localtime(&utc_time));
  fprintf(info_fp, "#created by cdda2wav "VERSION" %s\n"
	  "Track = %ld\n"
	  "Length = %ld\n"
	  "ISRC = %s\n"
#ifdef MD5_SIGNATURES
	  "MD-5(%d) = %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x\n"
#endif
	  , datetime
	  , track
	  , (nSamplesDone*40) /3	/* in milliseconds */
	  , g_toc[track-1].ISRC
#ifdef MD5_SIGNATURES
	  , md5blocksize
	  , MD5_result[0]
	  , MD5_result[1]
	  , MD5_result[2]
	  , MD5_result[3]
	  , MD5_result[4]
	  , MD5_result[5]
	  , MD5_result[6]
	  , MD5_result[7]
	  , MD5_result[8]
	  , MD5_result[9]
	  , MD5_result[10]
	  , MD5_result[11]
	  , MD5_result[12]
	  , MD5_result[13]
	  , MD5_result[14]
	  , MD5_result[15]
#endif
	  );
  fclose(info_fp);
  return 0;
}
#endif

/* set size and close audio file */
static void CloseAudio(char *fname_base, long int track, int bulk)
{
#ifdef INFOFILES
      /* define sizes in the header of the audio file */
      write_info_file(fname_base,track,nSamplesDone,bulk);
#endif

      /* define length */
      ExitAudioFile( audio, nSamplesDone*OutSampleSize*channels );

      close (audio);
      audio = -1;
}

/* On terminating:
 * define size-related entries in audio file header, update and close file */
static void CloseAll (void)
{
  int chld_return_status;
  int amichild;

  /* terminate child process first */
  amichild = child_pid == 0;

  if (amichild) {
      /* kill parent process */
/*      kill (getppid(), SIGINT);*/
      return;
  } else
      kill(child_pid,SIGINT);


  /* wait for child to terminate */
  wait(&chld_return_status);

  /* do general clean up */

#ifdef	ECHO_TO_SOUNDCARD
  if (echo) {
      close(soundcard_fd);
  }
#endif
  if (audio>=0) {
    CloseAudio(fname_base, current_track, bulk);
  }

  /* free shared memory and semaphores */
  free_semshm();

  /* switch to original mode and close device */
  if (cd_fd >= 0) {
      EnableCdda (FALSE);
      close(cd_fd);
      cd_fd = -1;
  }
}


/* report a fatal error, clean up and exit */
void FatalError (char *szMessage, ...)
{
  vfprintf (stderr, szMessage, (char *)(&szMessage + 1));
  exit (1);
}

/* request a shared memory block */
int shm_request(key_t key, unsigned int size, unsigned char **memptr)
{
  int   ret_val;
  int   shmflag;
  int   SHMEM_ID;
  int    cmd;
  struct shmid_ds buf;

  shmflag = IPC_CREAT | 0600;
  ret_val = shmget(key,size,shmflag);
  if ( ret_val == -1 )
  {
    perror("shmget");
    return -1;
  }

  SHMEM_ID = ret_val;
  cmd = IPC_STAT;
  ret_val = shmctl(SHMEM_ID,cmd,&buf);
#ifdef IPCTST
  fprintf(stderr, "%d: shmctl STAT= %d, SHM_ID: %d, key %ld cuid %d cgid %d mode %3o size %d\n",
          getpid(),ret_val,SHMEM_ID,
          buf.shm_perm.key,buf.shm_perm.cuid,buf.shm_perm.cgid,
          buf.shm_perm.mode,buf.shm_segsz);
#endif

  *memptr = shmat(SHMEM_ID, NULL, 0);
  if (*memptr == (unsigned char *) -1) {
    *memptr = NULL;
    return -1;
  }
  return SHMEM_ID;
}


/* open the audio output file and prepare the header. 
 * the header will be defined on terminating (when the size
 * is known). So hitting the interrupt key leaves an intact
 * file.
 */
static void OpenAudio (char *fname, unsigned long rate, long nBitsPerSample, 
		       long channels, unsigned long expected_bytes)
{
  uid_t my_euid = geteuid();
  uid_t my_uid  = getuid();

  if (audio == -1) {
    /* release superuser permissions */
    seteuid(my_uid);

    audio = open (fname, O_CREAT | O_WRONLY | O_TRUNC
#ifdef SYNCHRONOUS_WRITE
		       | O_SYNC
#endif
		  , 0600);
    if (audio == -1) {
      perror("open audio");
      FatalError ("Could not open %s\n", fname);
    }

    if (my_uid != my_euid)
      fchown(audio, my_uid, getgid());
  }
  InitAudioFile( audio, channels, rate, nBitsPerSample, expected_bytes );

#ifdef MD5_SIGNATURES
  if (md5blocksize)
    MD5Init (&context);
  md5count = md5blocksize;
#endif

  seteuid(my_euid);
}

long Remainder;
int quiet = 0;
int verbose = 3 + 16;


static void usage( void )
{
  fprintf( stderr,
"cdda2wav [-c chans] [-s] [-m] [-b bits] [-r rate] [-a divider]\n"
"         [-t track] [-i index] [-o offset] [-d duration]\n"
"         [-x] [-q] [-w] [-v] [-R] [-P] [-e] [-n sectors] [-N]\n"
"         [-D device] [-I interface] [-O audiotype] [-B] [audiofile]\n"
"         [-A auxdevice]\n"
"Version %s\n",VERSION );
  fprintf( stderr, "cdda2wav copies parts from audio cd's directly to wav or au files.\n"
"It requires a supported cdrom drive to work.\n"
"options: -D device   : set the cdrom or scsi generic device.\n"
"         -I interface: specify the interface for cdrom access.\n"
"                     : (generic_scsi or cooked_ioctl).\n"
"         -c channels : set 1 for mono, or 2 for stereo recording.\n"
"         -s          : set to stereo recording.\n"
"         -m          : set to mono recording.\n"
"         -x          : set to maximum quality (stereo/16-bit/44.1 KHz).\n"
"         -b bits     : set bits per sample per channel (8, 12 or 16 bits).\n"
"         -r rate     : set rate in samples per second. -R gives all rates\n"
"         -a divider  : set rate to 44100Hz / divider. -R gives all rates\n"
"         -R          : dump a table with all available sample rates\n"
"         -P          : switch off overlap sampling\n"
"         -n sectors  : read sectors per request.\n"
"         -t track    : select start track.\n"
"         -i index    : select start index.\n"
"         -o offset   : start 'offset' sectors behind start track/index.\n"
"                       one sector equivalents 1/75 second.\n"
"         -O audiotype: set wav or sun or raw audio format. Default is wav.\n"
"         -d duration : set recording time in seconds or 0 for whole track.\n"
"         -B          : record each track into a seperate file.\n"
"         -w          : wait for audio signal, then start recording.\n"
#ifdef	ECHO_TO_SOUNDCARD
"         -e          : echo audio data to sound device SOUND_DEV.\n"
#endif
"         -v level    : print informations on current cd (level: 0-31).\n"
"         -N          : no file operation.\n"
#ifdef MD5_SIGNATURES
"         -M count    : calculate MD-5 checksum for 'count' bytes.\n"
#endif
"         -q          : quiet operation, no screen output.\n"
#ifdef USE_GETOPT_LONG
"         -V          : print version information.\n"
"         -h          : this help screen.\n"  
#endif
"defaults: %s, %d bit, %g Hz, track 1, no offset, one track,\n"
"          type %s '%s', don't wait for signal, not quiet, not verbose,\n"
"          use %s, device %s, aux %s\n",
	  CHANNELS-1?"stereo":"mono", BITS_P_S, 44100.0 / UNDERSAMPLING, 
          AUDIOTYPE, FILENAME, DEF_INTERFACE, CD_DEVICE, AUX_DEVICE);
  fprintf( stderr, 
"parameters: (optional) a file name or - for standard output.\n");
  exit( 1 );
}


static unsigned long SectorBurst;
#if (SENTINEL > CD_FRAMESIZE_RAW)
#error block size for overlap check has not to be > sector size
#endif

/* and finally: the MAIN program */
long main ( long argc, char *argv [] )
{
  long lSector;
  long lSector_p1;
  long sector_offset = 0;
  long SamplesToWrite; 
  double time = 0.0;
  int track = 1;
  int index = 1;
  double rate = 44100 / UNDERSAMPLING;
  double int_part;
  int bits = BITS_P_S;
  char fname[200];
  int just_the_toc = 0;
  char int_name[100] = DEF_INTERFACE;
  char audio_type[10] = AUDIOTYPE;
  long i;
  unsigned char *p;
  int c;
  int long_option_index=0;

  /* Control those set-id privileges... */
  initsecurity();
  neverneedroot();     /* we don't have scsi_ioctl any more */

  /* command options parsing */
#ifdef USE_GETOPT_LONG
  while ( (c = getopt_long (argc,argv,optstring,options,&long_option_index)) 
		!= EOF)
#else
  while ( (c = getopt(argc, argv, "A:D:I:O:c:b:r:a:t:i:d:o:n:v:M:"
		                  "smweNqxPRB")) != EOF )
#endif /* USE_GETOPT_LONG */
	{
    switch (c) {
#ifdef MD5_SIGNATURES
      case 'M':
	md5blocksize = atoi( optarg );
        break;
#endif
      case 'D':    /* override device */
        strncpy( dev_name, optarg, sizeof(dev_name) );
        dev_name[sizeof(dev_name)-1]=0;
	break;
      case 'A':    /* override device */
        strncpy( aux_name, optarg, sizeof(aux_name) );
        aux_name[sizeof(aux_name)-1]=0;
	break;
      case 'I':    /* override interface */
	strcpy( int_name, optarg );
	break;
      case 'O':    /* override audio type */
	strncpy( audio_type, optarg, 4);
	break;
      case 'c':    /* override channels */
	channels = atoi( optarg );
	break;
      case 'b':    /* override bits */
	bits = atoi( optarg );
	break;
      case 'r':    /* override rate */
	rate = atoi( optarg );
	break;
      case 'a':    /* override rate */
	if (strtod( optarg, NULL ) != 0.0)
	    rate = fabs(44100.0 / strtod( optarg, NULL ));
	else {
	    fprintf(stderr, "-a requires a nonzero, positive divider.\n");
	    usage();
	}
	break;
      case 't':    /* override start track */
	track = atoi( optarg );
	break;
      case 'i':    /* override start index */
	index = atoi( optarg );
	break;
      case 'd':    /* override recording time */
	time = strtod( optarg, NULL );
	break;
      case 'o':    /* override offset */
	sector_offset = atol( optarg );
	break;
      case 'n':    /* read sectors per request */
	nsectors = atol( optarg );
	break;
      case 's':    /* stereo */
	channels = 2;
	break;
      case 'm':    /* mono */
	channels = 1;
	break;
      case 'x':    /* max */
	channels = 2; bits = 16; rate = 44100;
	break;
      case 'w':    /* wait for some audio intensity */
	waitforsignal = 1;
	break;
      case 'e':	   /* echo to sound acrd */
#ifdef	ECHO_TO_SOUNDCARD
	echo = 1;
#else
	fprintf(stderr, "There is no sound support compiled into %s.\n",argv[0]);
#endif
	break;	
      case 'v':    /* tell us more */
	verbose = atoi( optarg );
	break;
      case 'q':    /* be quiet */
	quiet = 1;
	verbose = 0;
	break;
      case 'N':    /* don't write to file */
	no_file = 1;
	break;
      case 'B':    /* bulk transfer */
	bulk = 1;
	if (time == 0.0)
	  time = 99999.0;
	break;
      case 'P':    /* prevent overlap reading */
	overlap = 0;
	break;
      case 'R':    /* list available rates */
	fprintf(stderr, "Available rates are:\n"
               "Rate   Divider      Rate   Divider      "
               "Rate   Divider      Rate   Divider\n");
	for (i = 1; i <= 44100 / 900; i++) {
	    fprintf(stderr, "%-7g  %2ld         %-7g  %2ld.5       ",
                    44100.0/i,i,44100/(i+0.5),i);
	    i++;
	    fprintf(stderr, "%-7g  %2ld         %-7g  %2ld.5\n",
                    44100.0/i,i,44100/(i+0.5),i);
	}
	exit(0);
#ifdef USE_GETOPT_LONG
	break; /* okay, not really needed for code, but good programming IMHO */
		case 'V':
			fprintf (stderr,"%s version %s\n",argv[0],VERSION);
			exit (0);
			break;
		case 'h':
			usage();
			break;
		case '?':
				fprintf (stderr,"use %s --help to get more information\n.",argv[0]);
				exit (1);
				break;
#endif
      default:
        usage();
    }
  }

  /* all options processed. Now a file name may follow */
  if ( optind < argc ) {
    if (!strcmp(argv[optind],"-")) {
      /* pipe mode */
      bulk = 0;
      audio = dup (fileno(stdout));
      strcpy( fname_base, "standard output" );
    } else {
      /* filename given */
      strncpy( fname_base, argv[optind], sizeof(fname_base)-8 );
      fname_base[sizeof(fname_base)-1]=0;
    }
  }

  /* check all parameters */
  if (verbose < 0 || verbose > 31) {
    fprintf(stderr, "Error: incorrect verbose level setting: %d\n",verbose);
    usage();
  }
  if (verbose == 0) quiet = 1;

  if ( channels != 1 && channels != 2 ) {
    fprintf(stderr, "Error: incorrect channel setting: %d\n",channels);
    usage();
  }

  if ( bits != 8 && bits != 12 && bits != 16 ) {
    fprintf(stderr, "Error: incorrect bits_per_sample setting: %d\n",bits);
    usage();
  }

  if ( rate < 827.0 || rate > 44100.0 ) {
    fprintf(stderr, "Error: incorrect sample rate setting: %g\n",rate);
    usage();
  }

  if ( modf( 2*44100.0 / rate, &int_part) >= 0.5 ) {
      int_part += 1.0;
      fprintf( stderr, "Nearest available sample rate is %g Hertz\n",
	      2*44100.0 / int_part);
  }
  Halved = ((int) int_part) & 1;
  rate = 2*44100.0 / int_part;
  undersampling = (int) int_part / 2.0;
  samples_to_do = undersampling;

  if ( time < 0 ) {
    fprintf(stderr, "Error: incorrect recording time setting: %f\n",time);
    usage();
  }

  if (!strcmp(int_name,"generic_scsi"))
      interface = GENERIC_SCSI;
  else if (!strcmp(int_name,"cooked_ioctl"))
      interface = COOKED_IOCTL;
  else  {
    fprintf(stderr, "Error: incorrect interface setting: %s\n",int_name);
    usage();
  }

#ifdef MD5_SIGNATURES
  if (md5blocksize < 0) {
    fprintf(stderr, "Error: incorrect i MD-5 count: %d\n",md5blocksize);
    usage();
  }
#endif

  if (no_file) fname_base[0] = '\0';

  /* setup interface and open cdrom device */
  /* request semaphores and shared memory */
  SetupInterface( int_name );

  /* get table of contents */
  tracks = ReadToc( g_toc );
  if ( verbose & (SHOW_TOC | SHOW_STARTPOSITIONS) )
    DisplayToc ();

  if (just_the_toc) exit(0);

  if (nsectors < 1+overlap) {
      fprintf(stderr, "Error: overlap (%d) is >= nsectors (%d)\n",overlap, nsectors);
      exit(-1);
  }

  /* check * init audio file */
  if (!strncmp(audio_type,"wav",3)) {
    InitAudioFile = InitWav;
    ExitAudioFile = ExitWav;
    GetSndHdrSize = GetWavHdrSize;
  } else if (!strncmp(audio_type, "sun", 3)) {
    InitAudioFile = InitSun;
    ExitAudioFile = ExitSun;
    GetSndHdrSize = GetSunHdrSize;
  } else if (!strncmp(audio_type, "raw", 3)) {
    InitAudioFile = InitRaw;
    ExitAudioFile = ExitRaw;
    GetSndHdrSize = GetRawHdrSize;
  } else {
    fprintf(stderr, "Error: incorrect audio type setting: %3s\n", audio_type);
    usage();
  }

  /* paranoia reigns supreme */
  sync();

  /* switch cdrom to audio mode */
  EnableCdda (TRUE);

  atexit ( CloseAll );

  if ( !FirstTrack () )
    FatalError ( "This disk has no audio tracks\n" );

  if ( verbose & (SHOW_MCN | SHOW_ISRC) )
    Read_MCN_ISRC();

  /* check if track is in range */
  if ( track < 1 || track > tracks ) {
    fprintf(stderr, "Error: incorrect track setting: %d\n",track);
    usage();
  }

  if (index != 1)
    sector_offset += ScanIndices( track, index );
  else if (verbose & SHOW_INDICES)
    ScanIndices( track, index );

  lSector = GetStartSector ( track );
  lSector_p1 = GetEndSector ( track ) + 1;

  if ( lSector < 0 )
    FatalError ( "track %d not found\n", track );

  lSector += sector_offset;
  /* check against end sector of track */
  if ( lSector >= lSector_p1 ) {
    fprintf( stderr, "sector offset exceeds track size (ignored)\n" );
    lSector -= sector_offset;
  }
  if ( time == 0.0 ) {
    /* set time to track time */
    nSamplesToDo = (lSector_p1 - lSector) * CD_FRAMESAMPLES;
    time = (lSector_p1 - lSector) / 75.0;
  } else {
    /* Prepare the maximum recording duration.
     * It is defined as the biggest amount of
     * adjacent audio sectors beginning with the
     * specified track/index/offset. */

    long lSector_p2 = GetLastSectorOnCd( track );
    if ( time > (lSector_p2 - lSector) / 75.0 ) {
      time = (lSector_p2 - lSector) / 75.0;
      lSector_p1 = lSector_p2;
    }

    /* calculate # of samples to read */
    nSamplesToDo = (long)(time*44100.0);
  }

  OutSampleSize = (1+bits/12);
  if (nSamplesToDo/undersampling == 0L) {
      fprintf( stderr, "time interval too short. Duration > %g secs!\n", 
	       undersampling/44100.0);
      exit(-2);
  }
  SamplesToWrite = nSamplesToDo*2/(int)int_part;

#ifdef	ECHO_TO_SOUNDCARD
  if (echo) {
    if ((soundcard_fd = open(SOUND_DEV, O_WRONLY, 0)) == EOF) {
        fprintf(stderr, "Cannot open "SOUND_DEV"\n");
        echo = 0;
    }
    { 
	int garbled_rate = rate;
	if (ioctl(soundcard_fd, SOUND_PCM_WRITE_RATE, &garbled_rate) == -1) {
	    fprintf(stderr, "Cannot set rate %g Hz for "SOUND_DEV"\n",rate);
	    echo = 0;
	}

	/* limited sound devices may not support stereo */
	if (ioctl(soundcard_fd, SOUND_PCM_WRITE_CHANNELS, &channels) == -1) {
	  int mono = 1;
	  fprintf(stderr, "Cannot set %d channel(s) for "SOUND_DEV"\n",channels);
	  if (ioctl(soundcard_fd, SOUND_PCM_WRITE_CHANNELS, &mono) == -1) {
	    fprintf(stderr, "Cannot set %d channel(s) for "SOUND_DEV"\n",mono);
	    echo = 0;
	  }
	}
	if (ioctl(soundcard_fd, SOUND_PCM_WRITE_BITS, &bits) == -1) {
	    fprintf(stderr, "Cannot set %d bits/sample for "SOUND_DEV"\n",bits);
	    echo = 0;
	}
	if ( abs(rate - garbled_rate) > rate / 20) {
	    fprintf(stderr, "sound device: next best sample rate is %d\n",garbled_rate);
	}
    }
  }
#endif

  if ( verbose & SHOW_SUMMARY ) {
      fprintf (stderr, "recording %.5f seconds %s with %d bits @ %g Hz"
	      ,time ,channels == 1 ? "mono":"stereo", bits, rate);
      if (*fname_base) fprintf(stderr, " ->'%s'...", fname_base );
      fputs("\n", stderr);
      if ( !waitforsignal )
	if (!bulk)
	  fprintf(stderr, "samplefile size will be %lu bytes.\n", GetSndHdrSize() +
	  SamplesToWrite*OutSampleSize*channels  ); 
	else {
	  int tracks_included;
	  tracks_included = GetTrack(lSector + nSamplesToDo/CD_FRAMESAMPLES -1) - track + 1;
	  fprintf(stderr, "samplefiles size total will be %lu bytes. %d tracks\n", 
		  tracks_included * GetSndHdrSize() +
	  SamplesToWrite*OutSampleSize*channels, tracks_included  ); 
	}
  }

  current_track = track;

  if ( !no_file ) {
    if (bulk)
      sprintf(fname, "%s.%02ld.%s",fname_base,current_track,audio_type);
    else
      strcpy(fname,fname_base);
    OpenAudio( fname, rate, bits, channels, SamplesToWrite*OutSampleSize*channels );
  }

  Remainder = (75 % nsectors)+1;

  /* Everything is set up. Now fork and let one process read cdda sectors
     and let the other one store them in a wav file */

  sh_bits = 16 - bits;		/* shift counter */

  i = nSamplesToDo;
  if (Halved && (i&1))
      i += 2;
  swap = 0;

  /* forking */
  child_pid = fork();

  if (child_pid < 0) {
    perror("fork()");
  }
  if (child_pid == 0) {
    /* child section */

    while (1) {

      /* switch buffers */
      if ( swap ) {
        p = bufferCdda + OFF + nsectors * CD_FRAMESIZE_RAW;
      } else {
        p = bufferCdda;
      }

      /* how many sectors should be read */
      if (overlap && nsectors*CD_FRAMESAMPLES > nSamplesToDo) {
	  SectorBurst = min(nsectors, 
			    overlap + ((i + CD_FRAMESAMPLES-1) / CD_FRAMESAMPLES));
      } else {
	  SectorBurst = min(nsectors, 
			    ((i + CD_FRAMESAMPLES-1) / CD_FRAMESAMPLES));
      }

      /* wait for buffer to be defined */
      semrequest(sem_id,swap);

      ReadCdRom( p, lSector, SectorBurst );

      /* Set buffer state to free */
      semrelease(sem_id,swap ^ 2);

      if (i >= nSamplesToDo)
	  i += overlap * CD_FRAMESAMPLES;
      i -= (SectorBurst - overlap) * CD_FRAMESAMPLES;
      if (i < 1) {

	  swap = (++swap & 1);
	  semrelease(sem_id,swap ^ 2);

	  /* the end, die */
	  exit(0);
      }

      lSector += SectorBurst - overlap;
      swap = !swap;
    }
    exit(0);
  } else {
    /* parent section */
    /* don't need these any more.  Good security policy says we get rid
       of them ASAP */
    neverneedroot();
    neverneedgroup();

    semrelease(sem_id,swap+1);

    while (nSamplesDone != SamplesToWrite) {

      /* switch buffers */
      if ( swap ) {
        p = bufferCdda + OFF + nsectors * CD_FRAMESIZE_RAW;
      } else {
        p = bufferCdda;
      }

      /* how many sectors should be read */
      SectorBurst = min(nsectors, 
         ((i + CD_FRAMESAMPLES-1) / CD_FRAMESAMPLES));

      /* set buffer state to free */
      semrelease(sem_id,swap);
      /* wait for buffer to be defined */
      semrequest(sem_id,swap ^ 2);

      /* when track end is reached, close current file and start a new one */

      while (lSector+SectorBurst > g_toc[current_track].dwStartSector) {
        if ( bulk && lSector < g_toc[current_track].dwStartSector ) {

#ifdef MD5_SIGNATURES
	  if (md5count) {
	    unsigned how_much = min(md5count, 
		    (g_toc[current_track].dwStartSector - lSector) * CD_FRAMESIZE_RAW);
	    MD5Update (&context, p, how_much);
	    md5count -= how_much;
	  }
#endif
	  if ( SaveBuffer ( p, channels,
	      (g_toc[current_track].dwStartSector - lSector) * CD_FRAMESAMPLES,
	      &nSamplesDone,
	      (g_toc[current_track].dwStartSector -
	       g_toc[current_track-1].dwStartSector) * CD_FRAMESAMPLES ) ) {
	    /* kill child */
	    kill(child_pid, SIGINT);
	    exit(2);
	  }

	  /* move residual samples upto buffer start */
	  memmove(p,
              p + (g_toc[current_track].dwStartSector - lSector)
                          * CD_FRAMESIZE_RAW,
              (SectorBurst - (g_toc[current_track].dwStartSector - lSector))
              * CD_FRAMESIZE_RAW);
	  i -= (g_toc[current_track].dwStartSector - lSector) * CD_FRAMESAMPLES;
	  SectorBurst -= g_toc[current_track].dwStartSector - lSector;
	  lSector = g_toc[current_track].dwStartSector;
	}

	if ( bulk && !no_file ) {
	  /* finish sample file for this track */
	  CloseAudio(fname_base, current_track, bulk);

	  if (verbose)
	    fprintf( stderr, "track %2ld successfully recorded\n", current_track);

	  /* build next filename */
	  sprintf(fname, "%s.%02ld.%s",fname_base,current_track+1,audio_type);
	  OpenAudio( fname, rate, bits, channels,
		    (g_toc[current_track].dwStartSector -
		     g_toc[current_track-1].dwStartSector)*CD_FRAMESIZE_RAW);
	}
	if (bulk) {
	  SamplesToWrite -= nSamplesDone;
	  nSamplesDone = 0;
	}
	current_track++;
	jitterShift = 0;
      }

      {
	int curr_limit = bulk ? min(i,(
                  g_toc[current_track].dwStartSector-lSector)*CD_FRAMESAMPLES)
			      : i;

#ifdef MD5_SIGNATURES
	if (md5count) {
	    unsigned how_much = min(md5count, min(curr_limit, SectorBurst * CD_FRAMESAMPLES)
		     * 4);
	    MD5Update (&context, p, how_much);
	    md5count -= how_much;
	}
#endif
	if ( SaveBuffer ( p, channels, min(curr_limit, SectorBurst * CD_FRAMESAMPLES), 
		       &nSamplesDone, SamplesToWrite ) ) {
	  /* kill child */
	  kill(child_pid, SIGINT);
	  exit(2);
	}
      }
      if (nSamplesDone + SkippedSamples >= SamplesToWrite) {
        int child_state;
	/* kill child */
	kill(child_pid, SIGINT);
        wait(&child_state);
	exit(2);
      }

      if (i == SamplesToWrite)
	  i -= overlap * CD_FRAMESAMPLES;
      i -= (SectorBurst - overlap) * CD_FRAMESAMPLES;
      if (i < 1) i = CD_FRAMESAMPLES;
      lSector += SectorBurst - overlap;
      swap = (++swap & 1);
    }
  }
  if (!quiet) fputs( "\n", stderr );

  return 0;
}
