/* resampling module 
 *
 * The audio data has been read. Here are the
 * functions to ensure a correct continuation
 * of the output stream and to convert to a
 * lower sample rate.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <linux/cdrom.h>

#include "cdda2wav.h"
#include "interface.h"


long waitforsignal = 0;	/* flag: wait for any audio response */

short undersampling;	/* conversion factor */
short samples_to_do;	/* loop variable for conversion */
int Halved;		/* interpolate due to non integral divider */

static long lsum = 0, rsum = 0;		/* accumulator for left/right channel */
static long ls2 = 0, rs2 = 0, ls3 = 0, rs3 = 0, auxl = 0, auxr = 0;

#define SYNC_SIZE	320L	/* has to be smaller than CD_FRAMESAMPLES */


static void init_search(const void *NEEDLE, size_t NEEDLE_LEN, size_t shift[])
{
  size_t *sp;
  size_t c;

  NEEDLE_LEN++;
  sp = shift;

  /* initialize jump table for characters non existent in NEEDLE */
  for ( c = UCHAR_MAX+1; c-- > 0; ) *sp++ = NEEDLE_LEN;

  /* initialize jump table for characters existent in NEEDLE */
  for ( ; --NEEDLE_LEN; shift[(int)*(unsigned char *)(NEEDLE++)] = NEEDLE_LEN );
}

void *memmem (const void *NEEDLE, size_t NEEDLE_LEN,
          const void *HAYSTACK, size_t HAYSTACK_LEN)
{
  const void *UPPER_LIMIT = HAYSTACK + HAYSTACK_LEN;

  static size_t shift[UCHAR_MAX+1];
  const unsigned char *cp;
  
  if (HAYSTACK_LEN < NEEDLE_LEN) return NULL;
  if (!HAYSTACK || !HAYSTACK_LEN) return NULL;
  if (!NEEDLE || !NEEDLE_LEN) return (void *) HAYSTACK;

  init_search(NEEDLE, NEEDLE_LEN, shift);

  cp = HAYSTACK + NEEDLE_LEN;
  while (cp <= (const unsigned char *) UPPER_LIMIT) {
    const unsigned char *p = NEEDLE;
    const unsigned char *t = cp - NEEDLE_LEN;

    do {
      if ((NEEDLE + NEEDLE_LEN) == p) {
	  return (void *) (cp - NEEDLE_LEN);
      }
    } while ( *p++ == *t++ );
    cp += shift[(int)*(cp)];
  }
  return NULL;
}

#if 0
/* A simple brute force approach */
void *memmem (const void *NEEDLE, size_t NEEDLE_LEN,
          const void *HAYSTACK, size_t HAYSTACK_LEN)
{
  const void *UPPER_LIMIT = HAYSTACK + HAYSTACK_LEN - NEEDLE_LEN;

  if (HAYSTACK_LEN < NEEDLE_LEN) return NULL;
  if (!HAYSTACK || !HAYSTACK_LEN) return NULL;
  if (!NEEDLE || !NEEDLE_LEN) return (void *) HAYSTACK;

  while (HAYSTACK <= UPPER_LIMIT) {
    if (!memcmp(NEEDLE, HAYSTACK, NEEDLE_LEN)) {
      return (void *) HAYSTACK;
    } else {
      HAYSTACK++;
    }
  }
  return NULL;
}
#endif

/* find continuation in new buffer */
static const void *sync_buffers(unsigned char *newbuf)
{
    const void *retval = newbuf;

    if (overlap) {
      /* find position of the last SYNC_SIZE bytes 
	 of the old buffer in the new buffer */
      unsigned char *oldbuf = bufferCdda + 
	           (1-swap) * (OFF + nsectors * CD_FRAMESIZE_RAW);

      /* compare the end of the previous buffer with the beginning
       * of the new one */
      const void *needle   = oldbuf+nsectors*CD_FRAMESIZE_RAW - SYNC_SIZE;
      const void *haystack = newbuf;
      size_t needle_len   = SYNC_SIZE;
      size_t haystack_len = overlap*CD_FRAMESIZE_RAW + SYNC_SIZE;

      retval = memmem(needle, needle_len, haystack, haystack_len);

      if (retval) {
          retval += SYNC_SIZE;
      }

    }
    return retval;
}


/* quadratic interpolation
 * p1, p3 span the interval 0 - 2. give interpolated value for 1/2 */
static long int interpolate( long int p1, long int p2, long int p3)
{
  return (3L*p1 + 6L*p2 - p3)/8L;
}

static int any_signal = 0;
static unsigned char *pStart;	/* running ptr defining 
				   end of output buffer */
static unsigned char *pDst;	/* start of output buffer */
/*
 * Write the filtered sample into the output buffer.
 */
static void emit_sample( long lsum, long rsum, long channels )
{
    /* output sample rate is reached. Now convert to output format */
    if ( channels == 1 ) {
	short sum;       /* mono section */
	sum = ( lsum + rsum ) >> (sh_bits + 1);
	if ( sh_bits == 8 ) {
	    if ( ( unsigned char ) sum != 0 ) {
		if ( !any_signal ) {
		    pStart = pDst;
		    any_signal = 1;
		}
	    }
	    *pDst++ = ( unsigned char ) sum + ( 1 << 7 );
	} else {
	    if ( sum != 0 ) {
		if ( !any_signal ) {
		    pStart = pDst;
		    any_signal = 1;
		}
	    }
	    if (lowendian) {
	      *( ( short * )pDst )++ = sum;
	    } else {
	      *pDst++ = ( unsigned char ) (sum >> 8);
	      *pDst++ = ( unsigned char ) sum;
	    }
	}
    } else {
	/* stereo section */
	lsum >>= sh_bits;
	rsum >>= sh_bits;
	if ( sh_bits == 8 ) {
	    if ( ( short ) lsum != 0 || 
		( short ) rsum != 0 ) {
		if ( !any_signal ) {
		    pStart = pDst;
		    any_signal = 1;
		}
	    }
	    *pDst++ = ( unsigned char )( short ) lsum + ( 1 << 7 );
	    *pDst++ = ( unsigned char )( short ) rsum + ( 1 << 7 );
	} else {
	    if ( ( short ) lsum != 0 || 
		( short ) rsum != 0) {
		if ( !any_signal ) {
		    pStart = pDst;
		    any_signal = 1;
		}
	    }
	    if (lowendian) {
	      *( ( short * )pDst )++ = ( short ) lsum;
	      *( ( short * )pDst )++ = ( short ) rsum;
	    } else {
	      *pDst++ = ( unsigned char ) (lsum >> 8);
	      *pDst++ = ( unsigned char ) lsum;
	      *pDst++ = ( unsigned char ) (rsum >> 8);
	      *pDst++ = ( unsigned char ) rsum;
	    }
	}
    }
}


#undef SHOW_JITTER
int jitterShift = 0; 

/* convert cdda data to required output format
 * sync code for panasonic cdroms included
 * 
 */
long SaveBuffer (unsigned char *p, long channels, 
                 long SamplesToDo, long *TotSamplesDone, long TotSamplesWanted)
{
  char *pSrc;		        /* start of cdrom buffer */
  char *pSrcStop;	        /* end of cdrom buffer */
  static int jitter = 0; 

  /* synchronisation code */

  if (*TotSamplesDone && overlap && SamplesToDo > CD_FRAMESAMPLES) {
    pSrc = (char *) sync_buffers(p);
    if ( !pSrc ) {
	fprintf(stderr, "no match: increase overlap (%u) and/or SYNC_SIZE (%ld).\n", overlap, SYNC_SIZE);
	fprintf(stderr, "no match: at position %ld from %ld. %ld todo\n",
		*TotSamplesDone, TotSamplesWanted, SamplesToDo);
	pSrc = p;
    }
    jitter = ((unsigned char *)pSrc - (p + overlap*CD_FRAMESIZE_RAW))/4;
    jitterShift += jitter;
    SamplesToDo -= jitter + overlap*CD_FRAMESAMPLES;
#if 0
    fprintf(stderr,
	    "Length: pre %d, diff1 %ld, diff2 %ld, min %ld\n", SamplesToDo,
	   (TotSamplesWanted - *TotSamplesDone),
	   SamplesNeeded((TotSamplesWanted - *TotSamplesDone), undersampling),
	   min(SamplesToDo, SamplesNeeded((TotSamplesWanted - *TotSamplesDone), undersampling)));
#endif
    SamplesToDo = min(SamplesToDo, SamplesNeeded((TotSamplesWanted - *TotSamplesDone), undersampling));

    p = pSrc;
  } else {
    pSrc = ( char * )p;
  }

  pDst = pSrc;
  pStart = pSrc;
  pSrcStop = ( char * )pSrc + SamplesToDo*4;
#ifdef SHOW_JITTER
    fprintf(stderr, "Total: %d, Jitter: %d\n", jitterShift, jitter);
#endif

  /* code for subsampling and output stage */

  /* optimize the case of no conversion */
  if ( undersampling == 1 && samples_to_do == 1 &&
       channels == 2 && OutSampleSize == 2 && !Halved) {
      /* output format is the original cdda format ->
       * just forward the buffer 
       */
      
      if ( waitforsignal && !any_signal ) {
	  while ((char *)pStart < pSrcStop && !*((long *)pStart)++);	/* scan for first signal */
	  if ( (char *)pStart != pSrcStop ) {
	      /* first non null amplitude is found in buffer */
	      any_signal = 1;
	  }
      }
      pDst = pSrcStop;		/* set pDst to end */
  } else {

#define none_missing	0
#define one_missing	1
#define two_missing	2
#define collecting	3

      static int sample_state = collecting;
      static int Toggle_on = 0;

    /* conversion required */
    while ( pSrc < pSrcStop ) {
	  
	long l,r;

	long iSamples_left = (pSrcStop - pSrc) / sizeof(signed short) / 2;
	/* LSB l, MSB l */
	l = *( ( signed short * )pSrc )++;	/* left channel */
	r = *( ( signed short * )pSrc )++;	/* right channel */

	switch (sample_state) {
	case two_missing:
two__missing:
	    ls2 += l; rs2 += r;
	    if (undersampling > 1) {
		ls3 += l; rs3 += r;
	    }
	    sample_state = one_missing;
	    break;
	case one_missing:
	    auxl = l; auxr = r;

	    ls3 += l; rs3 += r;
	    sample_state = none_missing;

	    /* FALLTHROUGH */
none__missing:
	case none_missing:
	    /* Filtered samples are complete. Now interpolate and scale. */

	    if (Halved && !Toggle_on) {
                lsum = interpolate(lsum, ls2, ls3)/(signed) undersampling;
	        rsum = interpolate(rsum, rs2, rs3)/(signed) undersampling;
            } else {
		lsum /= (signed) undersampling;
		rsum /= (signed) undersampling;
            }
	    emit_sample(lsum, rsum, channels);
	    /* reload counter */
	    samples_to_do = undersampling - 1;
	    lsum = auxl;
	    rsum = auxr;
	    /* reset sample register */
	    auxl = ls2 = ls3 = 0;
	    auxr = rs2 = rs3 = 0;
	    Toggle_on ^= 1;
	    sample_state = collecting;
	    break;
	case collecting:
	    if ( samples_to_do > 0) {
		samples_to_do--;
		if (Halved && !Toggle_on) {
		    /* Divider x.5 : we need data for quadratic interpolation */
		    iSamples_left--;

		    lsum += l; rsum += r;
		    if ( samples_to_do < undersampling - 1) {
			ls2 += l; rs2 += r;
		    }
		    if ( samples_to_do < undersampling - 2) {
			ls3 += l; rs3 += r;
		    }
		} else {
		    /* integral divider */
		    lsum += l;
		    rsum += r;
		    iSamples_left--;
		}
	    } else {
	        if (Halved && !Toggle_on) {
		    sample_state = two_missing;
		    goto two__missing;
		} else {
		    auxl = l;
		    auxr = r;
		    sample_state = none_missing;
		    goto none__missing;
		}
	    }
	    break;
	} /* switch state */

    } /* while */

    /* flush_buffer */
    if ((!samples_to_do && !Halved) ||
         (*TotSamplesDone + (pDst - pStart)/OutSampleSize/channels == TotSamplesWanted-1))
    {
	if (Halved && !Toggle_on) {
	    lsum = interpolate(lsum, ls2, ls3)/(signed) undersampling;
	    rsum = interpolate(rsum, rs2, rs3)/(signed) undersampling;
	} else {
	    lsum /= (signed) undersampling;
	    rsum /= (signed) undersampling;
	}
	emit_sample(lsum, rsum, channels);
	
	/* reload counter */
	samples_to_do = undersampling;
	
	/* reset sample register */
	lsum = auxl = ls2 = ls3 = 0;
	rsum = auxr = rs2 = rs3 = 0;
	Toggle_on ^= 1;
	sample_state = collecting;
    }

  } /* if optimize else */

  if ( !waitforsignal ) pStart = (unsigned char *)p;
  else if (any_signal) SkippedSamples += (pStart - p)/4;
  else SkippedSamples += (pSrcStop - (char *)p)/4;

  if ( !waitforsignal || any_signal ) {
    int retval = 0;
    int outlen = pDst - pStart;
    if (outlen <= 0) return 0;

#ifdef	ECHO_TO_SOUNDCARD
    if (echo) {
        write(soundcard_fd, pStart, outlen);
    }
#endif

    if ( no_file || 
	!( retval = write ( audio, pStart, outlen ) != outlen ) ) {
	outlen /= OutSampleSize*channels;
	*TotSamplesDone += outlen;
    } else {
      fprintf(stderr, "write(audio, 0x%p, %d)\n",pStart,outlen);
      perror("Probably disk space exhausted");
    }
    return retval;
  } else return 0;
}


